import re

from lxml import etree

EICR_MSG_ID_XPATH = "//hl7:ClinicalDocument/hl7:id"
RR_MSG_ID_XPATH = "//hl7:ClinicalDocument/hl7:section/hl7:id"

XML_PATH_DELIMITER = "/"

ECR_NAMESPACES = {
    "hl7": "urn:hl7-org:v3",
    "xsi": "http://www.w3.org/2005/Atom",
    "cda": "urn:hl7-org:v3",
    "sdtc": "urn:hl7-org:sdtc",
    "voc": "http://www.lantanagroup.com/voc",
}


def _get_xml_message_id(id_xml_tag: etree.Element) -> dict:
    # extracts the message id from the root and extension
    # attributes from the specified xml tag and returns
    # the results in a dictionary

    if id_xml_tag == []:
        return {}
    id_root = id_xml_tag[0].get("root")
    id_extension = id_xml_tag[0].get("extension")
    message_id = {"root": id_root, "extension": id_extension}
    return message_id


def get_ecr_message_ids(parsed_ecr) -> dict:
    # get the message ids for the eicr and the rr
    xml_eicr_id = _get_xml_message_id(
        parsed_ecr.xpath(EICR_MSG_ID_XPATH, namespaces=ECR_NAMESPACES)
    )
    xml_rr_id = _get_xml_message_id(
        parsed_ecr.xpath(RR_MSG_ID_XPATH, namespaces=ECR_NAMESPACES)
    )

    return {"eicr": xml_eicr_id, "rr": xml_rr_id}


def validate_xml_elements(xml_elements, config_field) -> list:
    """
    Matches the xml_elements to the config field requirements for
    values, attributes, and relative xml element requirements for
    values, and attributes. Returns list of matching fields.

    :param xml_elements: A list of xml elements
    :param config_field: A dictionary of the requirements of the field.
    :return: A list of matched xml elemnts
    """
    if not xml_elements:
        return []
    validated_elements = []
    for xml_element in xml_elements:
        if not _validate_xml_relatives(xml_element, config_field):
            continue
        found = _check_xml_names_and_attribs_exist(xml_element, config_field)
        if found:
            validated_elements.append(xml_element)
    return validated_elements


def _validate_xml_relatives(xml_element, config_field) -> bool:
    """
    Gets the 'relatives' portion of the configuration for a particular
    field and validates that the required attributes and values in the
    relative xml paths are present.

    :param xml_element: The key xml element that is being evaluated as
        well as its relative xml elements.
    :param config_field: A dictionary of the requirements for validating
        the key xml elements relative xml elements.
    :return: Bool - True if all relative xml elements match the criteria
        specified in the configuration, otherwise false.
    """
    relatives_valid = True
    relatives = config_field.get("relatives")
    if relatives is None:
        return relatives_valid
    for relative_config in relatives:
        base_cda_path = config_field.get("cdaPath")
        if (
            _validate_xml_related_element(
                xml_element=xml_element,
                cda_path=base_cda_path,
                relative_config=relative_config,
            )
            is None
        ):
            relatives_valid = False
    return relatives_valid


def _check_xml_names_and_attribs_exist(xml_element, config_field) -> bool:
    """
    Confirms that the xml element name and attributes specified in
    the configuration actually exist.

    :param xml_element: The key xml element that is being evaluated.
    :param config_field: A dictionary of the requirements for validating
        the key xml elements name and attributes
    :return: Bool - True if the name and attributes exist in the passed in
        xml element, otherwise false.
    """
    # If the configuration specified field name doesn't match the xml elements
    # tag name, then return false to go to the next xml element
    field_name = re.search(r"(?!\:)[a-zA-z]+\w$", config_field.get("cdaPath")).group(0)

    if field_name.lower() not in xml_element.tag.lower():
        return False

    # Don't try to evaluate matches for the xml element's attributes
    # if we are validating all fields as indicated in the configuration
    # for the field/xml element
    match_attributes = False if config_field.get("validateAll") == "True" else True
    if not match_attributes:
        return True

    # Check if the xml element passed in has the right attributes
    #  and if the values of the attribute correspond to the
    #  requirements in the configuration
    field_attributes = config_field.get("attributes")
    if field_attributes:
        for attribute in field_attributes:
            # If field is supposed to have an attribute and doesn't,
            # return that the field has failed validation for that
            # attribute.
            if not xml_element.get(attribute.get("attributeName")):
                return False
    else:
        # If there are not specified attributes to validate within the
        # configuration, but there are attributes in the xml element
        # then return a False/Fail
        # TODO: is this how this should be working??
        if xml_element.attrib:
            return False
    return True


def _validate_xml_related_element(xml_element, cda_path, relative_config) -> str:
    """
    Validates a related xml element, to the key xml element,
    against the configured rules for the related element's
    location to the key element and the existence of values
    and attributes for the related xml element.  If the relative
    xml element meets all the configured criteria, then the xml
    element will be returned, otherwise None.

    :param xml_element: The key xml element.
    :param cda_path: A string representing the location of the key xml element.
    :param config_field: A dictionary of the requirements for the related xml element.
    :return: A related xml element if the related xml element meets the configured
        criteria, otherwise return None.
    """
    if xml_element is None:
        return None
    relative_cda_path = relative_config.get("cdaPath")
    relative_tag_name = re.search(
        r"(?!\:)[a-zA-z]+\w$", relative_config.get("cdaPath")
    ).group(0)

    xml_iterator = _get_xml_relative_iterator(cda_path, relative_cda_path, xml_element)

    # if there are no related xml elements based upon the
    # specified cda paths then just return None
    if xml_iterator is None or len(xml_iterator) == 0:
        return None

    # sibling, descendant, child iterators
    # are of different type than ancestors
    # It's not clear why, but this prevents errors
    # and ensures that we have the correct element tag
    # and related element
    try:
        related_xml_element_tag = xml_iterator.tag
        related_xml_element = xml_iterator
    except Exception as e:
        print(e)
        related_xml_element = None
        related_xml_element_tag = None

    for xi in xml_iterator:
        # loop through iterator - unless it's
        # a ancestor - if that is the case then
        # the iterator itself IS the proper related
        # element and we want to use that - otherwise
        # loop through the list of elements in the list
        if related_xml_element_tag is None:
            related_xml_element = xi
            related_xml_element_tag = related_xml_element.tag

        if relative_tag_name in related_xml_element_tag:
            relative_validated = _check_xml_names_and_attribs_exist(
                xml_element=related_xml_element, config_field=relative_config
            )
            relative_attributes_validated = validate_xml_attributes(
                related_xml_element, relative_config
            )

            if not relative_validated or len(relative_attributes_validated) != 0:
                return None
            else:
                return related_xml_element
        else:
            related_xml_element = None
            related_xml_element_tag = None


def _get_xml_relative_iterator(cda_path, relative_cda_path, xml_element) -> list:
    """
    Gets an iterator, basically a list of xml elements, based on the key
    xml element's path and the relative xml element's path being
    searched for.

    :param cda_path: A string representing the location of the key xml element.
    :param relative_cda_path: A string representing the path of the relative
        xml element to be searched for.
    :param xml_element: The key xml element.
    :return: An lxml defined iterator of the relative xml elements
        based on the key xml element.
    """
    if xml_element is None:
        return None
    # get the difference of the number of '/' between the key cda_path and the
    # relative_cda_path
    diff = len(cda_path.split(XML_PATH_DELIMITER)) - len(
        relative_cda_path.split(XML_PATH_DELIMITER)
    )
    # element is on the same level of the main element
    if diff == 0:
        iter = []
        iter_forward = xml_element.itersiblings()
        iter_reverse = xml_element.itersiblings(preceding=True)
        for e in iter_forward:
            iter.append(e)
        for e in iter_reverse:
            iter.append(e)
        return iter
    # element is an acestor to the main element
    elif diff > 0:
        # get all the ancestors and put them into a list
        # and then get the one that is at the level
        # equal to the diff -1 (to account for array numbering)

        if len(list(xml_element.iterancestors())) < diff - 1:
            return None
        xml_relative_element = list(xml_element.iterancestors())[diff - 1]
        return xml_relative_element
    # element is a child of the main element
    elif diff == -1:
        return list(xml_element.iterchildren())
    # if diff is < -1 then it's a descendant and it's
    # going to have to return all tags under the base
    # xml element
    else:
        return list(xml_element.iterdescendants())


def validate_xml_attributes(xml_element, config_field) -> list:
    """
    Validates an xml element by checking if the configured attribute(s) exist
    and match the configured regex pattern for the value of the attribute.
    If the xml element, and it's subsequent attribute(s), do not pass the checks,
    based upon what is in the configuration, then an error message is added to
    the validation results.

    :param xml_element: The xml element that will have its attributes validated.
    :param config_field: A dictionary that contains the configuration that
        specifies what the attribute(s) for the xml element should be named
        and how the values should be patterned.
    :return: A list of errors or an empty list.  If the list is empty then
        the validation was successful.
    """
    config_attribs = config_field.get("attributes")
    if not config_attribs:
        return []

    error_messages = []
    for attribute in config_attribs:
        if "attributeName" in attribute:
            attribute_name = attribute.get("attributeName")
            attribute_value = xml_element.get(attribute_name)
            if not attribute_value:
                message = _get_ecr_custom_message(
                    config_field,
                    f"Could not find attribute '{attribute_name}'. "
                    + f"{get_xml_element_details(xml_element, config_field)}",
                )
                error_messages.append(message)
        if "regEx" in attribute:
            pattern = re.compile(attribute.get("regEx"))
            if (not attribute_value) or (not pattern.match(attribute_value)):
                message = _get_ecr_custom_message(
                    config_field,
                    f"Attribute: '{attribute_name}'"
                    + " not in expected format. "
                    + f"{get_xml_element_details(xml_element, config_field)}",
                )
                error_messages.append(message)
    return error_messages


def _get_ecr_custom_message(config_field, default_message):
    message = default_message
    if config_field.get("customMessage"):
        message = config_field.get("customMessage")
    return message


def get_xml_element_details(xml_element, config_field) -> str:
    """
    Gets the name, value, of the xml element referenced as well
    as the details of any configured relative xml elements for
    the referenced xml element and formats this information into
    a single string for use in the validation messages.

    :param xml_element: The key xml element.
    :param config_field: A dictionary of the configured requirements
        for the xml element.
    :return: A single string containing the information about the
        key xml element and its attribute(s).
    """
    if config_field is None or config_field == {}:
        return ""
    name = [f"Field name: '{config_field.get('fieldName')}'"]
    config_attributes = config_field.get("attributes")
    attributes = (
        ["Attributes:"] + _get_xml_attributes(xml_element, config_attributes)
        if config_attributes
        else []
    )
    relative_string = _get_xml_relatives_details(config_field.get("relatives"))
    value = (
        [f"value: '{''.join(xml_element.itertext())}'"]
        if (
            config_field.get("textRequired")
            and xml_element is not None
            and "".join(xml_element.itertext()) != ""
        )
        else []
    )
    return " ".join(name + value + attributes + relative_string)


def _get_xml_relatives_details(relatives_config: dict) -> list:
    """
    Gets the name and configured attribute information for
    any relative xml elements based upon the relative config
    section passed in and formats this information into
    a single string for use in the validation messages.

    :param relatives_config: A dictionary of the configured requirements
        for the relative xml element.
    :return: A single string, in a list, containing the information about the
        relative xml element and its attribute(s).
    """
    relative_string = []
    if relatives_config:
        relative_string.append("Related elements:")
        for rel_config in relatives_config:
            relative_field_name = rel_config.get("name")
            relative_name = (
                ["Field name: '" + relative_field_name + "'"]
                if relative_field_name
                else []
            )
            config_related_attributes = rel_config.get("attributes")
            relative_attributes = (
                ["Attributes:"] + _get_xml_attributes(None, config_related_attributes)
                if config_related_attributes
                else []
            )
            relative_string += relative_name
            relative_string += relative_attributes
    return relative_string


def _get_xml_attributes(xml_element, config_attributes) -> list:
    """
    Takes the xml element, along with its configuration to create
    a list of strings that include attribute name, required
    patterns, and the actual value of said attribute.

    :param xml_element: The key xml element.
    :param config_attributes: A dictionary of the configured requirements
        for the attributes for the key xml element.
    :return: A single string containing the information about the
        key xml elements attribute(s).
    """
    attrs = []
    if config_attributes is None or config_attributes == "":
        return attrs
    for attribute in config_attributes:
        attr_index = len(attrs) + 1
        attribute_name = attribute.get("attributeName")
        reg_ex = attribute.get("regEx")
        reg_ex_string = (
            f" with the required value pattern: '{reg_ex}'" if reg_ex else ""
        )

        if xml_element is not None and xml_element.get(attribute_name):
            attribute_value = f" actual value: '{xml_element.get(attribute_name)}'"
        else:
            attribute_value = ""

        attrs.append(
            f"attribute #{attr_index}: "
            + f"'{attribute_name}'{reg_ex_string}{attribute_value}"
        )
    return [", ".join(attrs)]


def validate_xml_value(xml_element, config_field) -> list:
    """
    Validates the value of an xml element (ie... between the tags) based upon
    validating the xml element exists within the relative xml element location
    based upon the configuration of the parent of the xml element.
    Then it validates that the value of xml element matches the value pattern
    listed in the config.

    :param xml_element: The key xml element being evaluated.
    :param config_field: A dictionary that contains the configuration information
        necessary to validate the location and value of the key xml element
    :return: A list of error messages if validation fails, otherwise an
        empty list.
    """
    # If value for xml element is not required return empty list
    if not config_field.get("textRequired"):
        return []

    value = "".join(xml_element.itertext())
    config_regex = config_field.get("regEx")
    # first check if the value matches the regex from the config
    if config_regex is not None:
        pattern = re.compile(config_regex)
        if pattern.match(value) is None:
            message = _get_ecr_custom_message(
                config_field,
                "The field value does not exist or "
                + "doesn't match the following pattern: '"
                + config_regex
                + f"'. For the {get_xml_element_details(xml_element, config_field)}",
            )
            return [message]
        else:
            return []
    # otherwise just verify that a value is present
    else:
        if value is not None and value != "":
            return []
        else:
            message = _get_ecr_custom_message(
                config_field,
                "Field does not have a value. "
                + f"{get_xml_element_details(xml_element, config_field)}",
            )
            return [message]



import csv
import importlib.resources
import json
import os
import pathlib
import sqlite3 as sql
from typing import List
from typing import Literal
from typing import Union

import pyarrow as pa
import pyarrow.parquet as pq
import yaml
from jsonschema import validate


def load_schema(path: pathlib.Path) -> dict:
    """
    Given the path to a local YAML or JSON file containing a schema,
    loads the file and return the resulting schema as a dictionary.
    If the file can't be found, raises an error.

    :param path: The file path to a YAML file holding a schema.
    :raises ValueError: If the provided path points to an unsupported file type.
    :raises FileNotFoundError: If the file to be loaded could not be found.
    :raises JSONDecodeError: If a JSON file is provided with invalid JSON.
    :return: A dict representing a schema read from the given path.
    """
    try:
        with open(path, "r") as file:
            if path.suffix == ".yaml":
                schema = yaml.safe_load(file)
            elif path.suffix == ".json":
                schema = json.load(file)
            else:
                ftype = path.suffix.replace(".", "").upper()
                raise ValueError(f"Unsupported file type provided: {ftype}")
        validate_schema(schema)
        return schema
    except FileNotFoundError:
        raise FileNotFoundError(
            "The specified file does not exist at the path provided."
        )
    except json.decoder.JSONDecodeError as e:
        raise json.decoder.JSONDecodeError(
            "The specified file is not valid JSON.", e.doc, e.pos
        )


def validate_schema(schema: dict):
    """
    Validates the schema structure, ensuring all required schema elements are present
    and all schema elements are of the expected data type.

    :param schema: A declarative, user-defined specification, for one or more tables,
        that defines the metadata, properties, and columns of those tables as they
        relate to FHIR resources.
    :raises jsonschema.exception.ValidationError: If the schema is invalid.
    """
    # Load validation schema
    with importlib.resources.open_text(
        "phdi.tabulation", "validation_schema.json"
    ) as file:
        validation_schema = json.load(file)

    validate(schema=validation_schema, instance=schema)


def write_data(
    tabulated_data: List[List],
    directory: str,
    output_type: Literal["csv", "parquet", "sql"],
    filename: str = None,
    db_file: str = None,
    db_tablename: str = None,
    pq_writer: pq.ParquetWriter = None,
    schema: dict = None,
    table_name: str = None,
) -> Union[None, pq.ParquetWriter]:
    """
    Writes a set of tabulated data to a particular output format on disk
    (one of CSV, Parquet, or SQL). For CSV and Parquet writing, a filename
    must be provided to write output to. In the case of the SQL format,
    a database file (if one exists and is being modified) must be specified
    along with a table name in place of a filename. Creates new data files
    if the given options specify a file that doesn't exist, and appends
    data to already-present files if they do.

    :param tabulated_data: A list of lists in which the first element
      is the headers for the table-to-write and subsequent elements
      are rows in the table.
    :param directory: The directory in which to write the output data
      (if `output_type` is CSV or Parquet), or the directory in which
      the `db_file` is stored (if a database already exists) or should
      be created in (if a database does not already exist).
    :param output_type: The format the data should be written to.
    :param filename: The name of the CSV or Parquet file data should be
      written to. Omit if `output_type` is SQL. Default: `None`.
    :param db_file: The name of the database file to either create or
      access when writing to a SQL format. Omit if `output_type` is not
      SQL. Default: `None`.
    :param db_tablename: The name of the table in the database to create
      or write data to. Omit if `output_type` is not SQL. Default: `None`.
    :param pq_writer: A pre-existing `ParquetWriter` object that can be
      used to append data to a parquet format. Used in cases where
      incremental writing to a parquet destination is desired. Omit if
      `output_type` is not Parquet. Default: `None`.
    """
    # some elements may themselves contain lists, if selection_criteria = all is used
    for i in range(1, len(tabulated_data)):
        table_list = tabulated_data[i]
        for row, elt in enumerate(table_list):
            if isinstance(elt, list):
                tabulated_data[i][row] = _convert_list_to_string(elt)
            if isinstance(elt, dict):
                tabulated_data[i][row] = str(elt)

    if output_type == "csv":
        write_headers = (
            False if os.path.isfile(os.path.join(directory, filename)) else True
        )
        with open(os.path.join(directory, filename), "a", newline="") as fp:
            writer = csv.writer(fp, dialect="excel")
            if write_headers:
                writer.writerow(tabulated_data[0])
            writer.writerows(tabulated_data[1:])

    if output_type == "parquet":
        if schema and table_name:
            # Create the parquet schema based on the config file
            pq_schema = _create_pa_schema_from_table_schema(
                schema, tabulated_data[0], table_name
            )
        else:
            pq_schema = None
        # Rearrange data so that it is column, not row based as parquet needs
        parquet_data = _create_parquet_data(tabulated_data, pq_schema)
        if pq_schema:
            table = pa.Table.from_arrays(
                _create_from_arrays_data(parquet_data[1:]), schema=pq_schema
            )
        else:
            # If no schema file is provided, use the names field
            table = pa.Table.from_arrays(
                _create_from_arrays_data(parquet_data[1:]), names=tabulated_data[0]
            )
        if pq_writer is None:
            pq_writer = pq.ParquetWriter(
                os.path.join(directory, filename), table.schema
            )
        pq_writer.write_table(table=table)
        return pq_writer

    # @TODO:
    # 1. support username and passwords for database access
    # 2. figure out a more intelligent way to serialize data that's being
    # written to the SQL DB; there's the possibility that we need to write
    # list data to the table if an anchor has multiple reverse references,
    # but SQL can't write native python lists, so we need to either
    # stringify them, json.dumps serialize them, or binarize them
    if output_type == "sql":
        # We need a file-space cursor to operate on a connected table
        conn = sql.connect(os.path.join(directory, db_file))
        cursor = conn.cursor()

        # Create requested table if it's not already in the database
        headers = tuple(tabulated_data[0])
        cursor.execute(f"CREATE TABLE IF NOT EXISTS {db_tablename} {headers};")

        # Format the programmatic values correctly by number of cols
        # Need this ugly construction so that the binding placeholder qmarks
        # don't have quotes around them individually when parsed by the SQL
        # engine
        hdr_placeholder = "({})".format(", ".join("?" * len(tabulated_data[0])))
        tuple_data = [tuple([str(x) for x in row]) for row in tabulated_data[1:]]
        # Have to do a format string this way to avoid an empty f-string error
        print(tuple_data)
        cursor.executemany(
            "INSERT INTO {} {} VALUES {};".format(
                db_tablename, headers, hdr_placeholder
            ),
            tuple_data,
        )

        conn.commit()
        conn.close()


def _convert_list_to_string(val: list) -> str:
    """
    Serializes a given list into a string, separating values with commas.

    :param val: A list that may contain other types of objects to be stringified
      for use in CSV, SQL, etc
    :return: A string representation of the list
    """
    for i, v in enumerate(val):
        if isinstance(v, list):
            val[i] = _convert_list_to_string(v)
        elif isinstance(v, dict):
            val[i] = str(v)
        elif not isinstance(v, str):
            val[i] = str(v)
    return (",").join(val)


def _create_pa_schema_from_table_schema(
    schema: dict, col_names: List, table_name: str
) -> pa.Schema:
    """
    Returns a parquet schema based on the schema definition file provided to the
      function. Defaults to String.

    :param schema: A dict value that is defined by the user which contains the structure
      of the data.
    :param col_names: A list of column names that the parquet schema is being generated
      for.
    :param table_name: A string of the table name that the parquet schema is being
      generated for.
    :return: A pyarrow schema object based on the schema of the table, column names, and
      which table is being used.
    """
    table_columns = schema["tables"][table_name]["columns"]
    pa_schema_arr = []
    for name in col_names:
        if name not in table_columns:
            pa_schema_arr.append((name, pa.string()))
            continue
        for col in table_columns:
            if str(col) == str(name):
                data_type = (
                    str(table_columns[col]["data_type"])
                    if "data_type" in table_columns[col]
                    else False
                )

                if data_type == "number":
                    pa_schema_arr.append(pa.field(name, pa.float32()))
                elif data_type == "boolean":
                    pa_schema_arr.append(pa.field(name, pa.bool_()))
                else:
                    pa_schema_arr.append(pa.field(name, pa.string()))
    pa_schema = pa.schema(pa_schema_arr)
    return pa_schema


def _create_from_arrays_data(row_data: List) -> List:
    """
    Returns a list that is one array per column. Accepts list that is one
      array per row.
    :param row_data: A list that is made of multiple arrays
    :return: A list of lists.
    """
    if len(row_data) == 0:
        return row_data
    col_data = []
    for row in row_data:
        for i, data in enumerate(row):
            if (len(col_data) - 1) < i:
                col_data.append([])
            col_data[i].append(data)
    return col_data


def _create_parquet_data(data: List[List], pq_schema: pa.Schema) -> List[List]:
    """
    Returns a list with the data being modified to the data types specified by the
      pyarrow schema.
    :param data: A list of lists with multiple objects.
    :param pq_schema: A pyarrow schema file which references the data in the data file.
    :return: A list of lists with data types specified in the pyarrow schema.
    """
    if pq_schema is None:
        for row in data[1:]:
            for i, elm in enumerate(row):
                if elm is None:
                    row[i] = ""
                else:
                    row[i] = str(elm)
        return data
    for row in data[1:]:
        for i, elm in enumerate(row):
            data_type_index = pq_schema.get_field_index(data[0][i])
            data_type = pq_schema.types[data_type_index]
            if data_type == "string":
                if isinstance(elm, str):
                    row[i] = elm
                else:
                    row[i] = str(elm)
            elif data_type == "float":
                if isinstance(elm, float):
                    row[i] = elm
                else:
                    row[i] = float(elm)
            elif data_type == "bool_":
                if isinstance(elm, bool):
                    row[i] = elm
                else:
                    row[i] = bool(elm)
            else:
                row[i] = elm
    return data



from typing import Literal
from typing import Union

import requests

from phdi.geospatial.core import BaseGeocodeClient
from phdi.geospatial.core import GeocodeResult
from phdi.transport import http_request_with_retry


class CensusGeocodeClient(BaseGeocodeClient):
    """
    Implementation of a geocoding client using the Census API.
    """

    def __init__(self):
        self.__client = ()

    def geocode_from_str(self, address: str) -> Union[GeocodeResult, None]:
        """
        Geocodes a string-formatted address using Census API with searchtype =
        "onelineaddress". If a result is found, encodes as a GeocodeResult object and
        return, otherwise the return None.

        :param address: The address to geocode, given as a string.
        :param searchtype: onelineaddress OR address # doesn't yet support coordinates.
        :raises ValueError: If address does not include street number and name.
        :return: A standardized address enriched with lat, lon, census tract, and more.
            Returns None if no valid result.
        """
        # Check for street num and name at minimum
        if address == "":
            raise ValueError("Address must include street number and name at a minimum")

        formatted_address = self._format_address(address, searchtype="onelineaddress")
        url = self._get_url(formatted_address)
        response = self._call_census_api(url)

        return self._parse_census_result(response)

    def geocode_from_dict(self, address: dict) -> Union[GeocodeResult, None]:
        """
        Geocodes the provided address, which is formatted as a dictionary.
        using the Census API with searchtype = "address". If a result is found, encodes
        as a GeocodeResult object and return, otherwise returns None.

        The given dictionary should conform to standard nomenclature around address
        fields, including:

        * `street`: the number and street address
        * `street2`: additional street level information (if needed)
        * `apartment`: apartment or suite number (if needed)
        * `city`: city to geocode
        * `state`: state to geocode
        * `postal_code`: the postal code to use
        * `urbanization`: urbanization code for area, sector, or regional
        * `development`: (only used for Puerto Rican addresses)

        Street must be included to use this function; however, a minimum of street,
        city, and state are suggested for the best matches.

        :param address: A dictionary with fields outlined above.
        :raises ValueError: If address does not include street number and name.
        :return: A standardized address enriched with lat, lon, census tract, and more.
            Returns None if no valid result.
        """

        # Check for street num and name at minimum
        if address.get("street", "") == "":
            raise ValueError("Address must include street number and name at a minimum")

        # Configure the lookup with whatever provided address values
        # were in the user-given dictionary
        formatted_address = self._format_address(address, searchtype="address")
        url = self._get_url(formatted_address)
        response = self._call_census_api(url)

        return self._parse_census_result(response)

    @staticmethod
    def _format_address(
        address: Union[str, dict], searchtype: Literal["onelineaddress", "address"]
    ):
        """
        Formats an address for Census API call according to the given address type. A
        single line address, e.g, "100 5th Ave New York, NY" uses the "onelineaddress"
        searchtype while a dictionary formatted address uses the "address" searchtype.

        :param address: The address to geocode, given as a string or dictionary.
        :param searchtype: onelineaddress OR address.
        :raises ValueError: If address cannot be geocoded with specificity because it
            does not include city, state, and/or zipcode.
        :return: A properly formatted address for the Census API call, given as a
            string.
        """
        # Check that the address contains structure number and street name
        if searchtype == "onelineaddress":
            address = address.replace(" ", "+").replace(",", "%2C")
            return f"onelineaddress?address={address}"
        elif searchtype == "address" and isinstance(address, dict):
            street = address.get("street", "").replace(" ", "+").replace(",", "%2C")
            city = address.get("city", "").replace(" ", "+").replace(",", "%2C")
            state = address.get("state", "").replace(" ", "+").replace(",", "%2C")
            zip = address.get("zip", "").replace(" ", "+").replace(",", "%2C")

            # If only "street" is present, format address with
            # searchtype = "onelineaddress"
            if any(element != "" for element in [city, state, zip]):
                # Add non-empty elements
                formatted_address = f"address?street={street}"
                for element in [city, state, zip]:
                    if element == "":
                        continue
                    else:
                        if element == city:
                            formatted_address += f"&city={city}"
                        elif element == state:
                            formatted_address += f"&state={state}"
                        elif element == zip:
                            formatted_address += f"&zip={zip}"
                return formatted_address

            else:
                return f"onelineaddress?address={street}"

    @staticmethod
    def _get_url(address: str):
        """
        Gets URL for Census API given inputs.

        :param address: The formatted address to geocode, given as a string.
        :return: A URL for the Census API request, as a string.
        """
        url = (
            f"https://geocoding.geo.census.gov/geocoder/geographies/{address}"
            + "&benchmark=Public_AR_Census2020"
            + "&vintage=Census2020_Census2020"
            + "&layers=[10]"
            + "&format=json"
        )
        return url

    @staticmethod
    def _call_census_api(url):
        """
        Calls the Census endpoint with a given URL using the http_request_with_retry
        method.

        :param url: A URL for the Census API request, as a string.
        :raises requests.HTTPError: If an unexpected status code is returned.
        :return: A response from queried endpoint.
        """
        http_url = url
        http_action = "GET"
        http_header = {"some-header": "some-header-value"}
        http_retry_count = 5

        response = http_request_with_retry(
            http_url,
            http_retry_count,
            http_action,
            [http_action],
            http_header,
        )

        if response.status_code != 200:
            raise requests.HTTPError(response=response)
        else:
            return response.json()["result"]

    @staticmethod
    def _parse_census_result(lookup) -> Union[GeocodeResult, None]:
        """
        Parses a returned Census geocoding result into our standardized GeocodeResult
        class. If the Census lookup is null or doesn't include matched address
        information, returns None instead.

        :param response: The Census API client instantiated for geocoding.
        :return: A parsed and standardized address enriched with lat, lon, census tract,
             and more. Returns None if no valid result.
        """

        if lookup is not None and lookup.get("addressMatches"):
            addressComponents = lookup.get("addressMatches", [{}])[0].get(
                "addressComponents", {}
            )
            blockComponents = (
                lookup.get("addressMatches", [{}])[0]
                .get("geographies", {})
                .get("Census Blocks", [None])[0]
            )
            tractComponents = (
                lookup.get("addressMatches", [{}])[0]
                .get("geographies", {})
                .get("Census Tracts", [None])[0]
            )
            countyComponents = (
                lookup.get("addressMatches", [{}])[0]
                .get("geographies", {})
                .get("Counties", [None])[0]
            )
            coordinateComponents = lookup.get("addressMatches", [{}])[0].get(
                "coordinates", {}
            )

            # Format the Census result into our standard dataclass object
            return GeocodeResult(
                line=[
                    lookup["addressMatches"][0]["matchedAddress"].split(",")[0].strip()
                ],
                city=addressComponents.get("city", ""),
                state=addressComponents.get("state", ""),
                postal_code=addressComponents.get("zip", ""),
                county_fips=blockComponents.get("STATE", "")
                + blockComponents.get("COUNTY", ""),
                county_name=countyComponents.get("BASENAME", ""),
                lat=coordinateComponents.get("y", None),
                lng=coordinateComponents.get("x", None),
                geoid=blockComponents.get("GEOID", ""),
                census_tract=tractComponents.get("BASENAME", ""),
                census_block=blockComponents.get("BASENAME", ""),
            )



import copy

from phdi.fhir.utils import find_entries_by_resource_type
from phdi.fhir.utils import get_field
from phdi.fhir.utils import get_one_line_address
from phdi.linkage.link import generate_hash_str


def add_patient_identifier_in_bundle(
    bundle: dict, salt_str: str, overwrite: bool = True
) -> dict:
    """
    Given a FHIR resource bundle:

    * Identifies all patient resource(s) in the bundle
    * Adds the hash string to the list of identifiers held in that patient resource

    :param bundle: The FHIR bundle for whose patients to add a linking identifier.
    :param salt_str: The salt to use with the hash. This is intended to prevent
      reverse engineering of the PII used to create the hash.
    :param overwrite: If true, `bundle` is modified in-place;
      if false, a copy of `bundle` modified and returned.  Default: `True`
    :return: The bundle, resources updated with additional patient identifier.
    """
    if not overwrite:
        bundle = copy.deepcopy(bundle)

    for entry in find_entries_by_resource_type(bundle, "Patient"):
        patient = entry.get("resource")
        add_patient_identifier(patient, salt_str)
    return bundle


def add_patient_identifier(
    patient_resource: dict, salt_str: str, overwrite: bool = True
):
    """
    Given a FHIR Patient resource:

    * Extracts name, DOB, and address information
    * Computes a unique hash string based on these fields
    * Adds the hash string to resource

    :param patient_resource: The FHIR patient resource to add a linking identifier.
    :param salt_str: The salt to use with the hash. This is intended to prevent
      reverse engineering of the PII used to create the hash.
    :param overwrite: If true, `patient_resource` is modified in-place;
      if false, a copy of `patient_resource` modified and returned.  Default: `True`
    :return: The resource updated with additional patient identifier.
    """
    if not overwrite:
        patient_resource = copy.deepcopy(patient_resource)

    # Combine given and family name
    recent_name = (
        get_field(patient_resource, field="name", use="official", require_use=False)
        or {}
    )
    name_parts = recent_name.get("given", []) + [recent_name.get("family", "")]
    name_str = "-".join([n for n in name_parts if n])

    address_line = ""
    if "address" in patient_resource:
        address = (
            get_field(patient_resource, field="address", use="home", require_use=False)
            or {}
        )
        address_line = get_one_line_address(address)

    # TODO Determine if minimum quality criteria should be included, such as min
    # number of characters in last name, valid birth date, or address line
    # Generate and store unique hash code
    link_str = f'{name_str}-{patient_resource["birthDate"]}-{address_line}'
    hashcode = generate_hash_str(link_str, salt_str)

    if "identifier" not in patient_resource:
        patient_resource["identifier"] = []

    # TODO Follow up on the validity and source of the comment about the system
    # value corresponding to the FHIR specification. Need to either add a citation
    # or correct the wording to more properly reflect what it represents.
    patient_resource["identifier"].append(
        {
            "value": hashcode,
            # Note: this system value corresponds to the FHIR specification
            # for a globally used / generated ID or UUID--the standard here
            # is to make the use "temporary" even if it's not
            "system": "urn:ietf:rfc:3986",
            "use": "temp",
        }
    )
    return patient_resource



from phdi.fhir.transport.export import export_from_fhir_server
from phdi.fhir.transport.http import fhir_server_get
from phdi.fhir.transport.http import http_request_with_reauth
from phdi.fhir.transport.http import upload_bundle_to_fhir_server

__all__ = [
    "http_request_with_reauth",
    "fhir_server_get",
    "upload_bundle_to_fhir_server",
    "export_from_fhir_server",
]



from phdi.fhir.harmonization.standardization import double_metaphone_bundle
from phdi.fhir.harmonization.standardization import double_metaphone_patient
from phdi.fhir.harmonization.standardization import standardize_dob
from phdi.fhir.harmonization.standardization import standardize_names
from phdi.fhir.harmonization.standardization import standardize_phones

__all__ = (
    "double_metaphone_bundle",
    "double_metaphone_patient",
    "standardize_names",
    "standardize_phones",
    "standardize_dob",
)



import json
from datetime import datetime
from datetime import timezone
from typing import List
from typing import Literal
from typing import Union

from azure.core.credentials import AccessToken
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from azure.storage.blob import BlobServiceClient
from azure.storage.blob import ContainerClient

from phdi.cloud.core import BaseCloudStorageConnection
from phdi.cloud.core import BaseCredentialManager


class AzureCredentialManager(BaseCredentialManager):
    """
    Defines a credential manager used for connecting to Azure.
    """

    @property
    def resource_location(self) -> str:
        return self.__resource_location

    @property
    def scope(self) -> str:
        return self.__scope

    @property
    def access_token(self) -> AccessToken:
        return self.__access_token

    def __init__(self, resource_location: str = None, scope: str = None):
        """
        Creates a new AzureCredentialManager object.

        :param resource_location: The URL or other location of the requested resource.
        :param scope: A space-delimited list of scopes to limit access to resource.
          Default: `None`
        """
        self.__resource_location = resource_location
        self.__scope = scope
        self.__access_token = None

        if self.scope is None:
            self.__scope = f"{self.resource_location}/.default"

    def get_credential_object(self) -> object:
        """
        Gets an Azure-specific credential object.

        :return: An instance of one of the \\*Credential objects from the
          `azure.identity` package.
        """
        return DefaultAzureCredential()

    def get_access_token(self, force_refresh: bool = False) -> str:
        """
        Obtains an access token from the Azure identity provider. Returns the
        access token string, refreshed if expired or force_refresh is specified.

        :param force_refresh: `True` if a new token should be requested, regardless
          of expiration timestamp. `False` otherwise. Default: `False`
        :return: An Azure access token.
        """
        if force_refresh or (self.access_token is None) or self._need_new_token():
            creds = self.get_credential_object()
            self.__access_token = creds.get_token(self.scope)

        return self.access_token.token

    def _need_new_token(self) -> bool:
        """
        Determines whether the token already stored for this object can be reused,
        or if it needs to be re-requested. A new token is needed if a token has not
        yet been created, or if the current token has expired.

        :return: True if a new Azure access token is needed; false otherwise.
        """
        try:
            current_time_utc = datetime.now(timezone.utc).timestamp()
            return self.access_token.expires_on < current_time_utc
        except AttributeError:
            # access_token not set
            return True

    def get_secret(self, secret_name: str, key_vault_name: str) -> str:
        """
        Get the value of a secret from an Azure key vault given the names of the vault
        and the secret.

        :param secret_name: The name of the secret whose value should be retrieved from
            the key vault.
        :param key_vault_name: The name of the key vault where the secret is stored.
        :return: The value of the secret specified by secret_name.
        """

        vault_url = f"https://{key_vault_name}.vault.azure.net"
        secret_client = SecretClient(
            vault_url=vault_url, credential=self.get_credential_object()
        )
        return secret_client.get_secret(secret_name).value


class AzureCloudContainerConnection(BaseCloudStorageConnection):
    """
    Defines a connection used for interacting with cloud storage in Azure.
    """

    @property
    def storage_account_url(self) -> str:
        return self.__storage_account_url

    @property
    def cred_manager(self) -> AzureCredentialManager:
        return self.__cred_manager

    def __init__(self, storage_account_url: str, cred_manager: AzureCredentialManager):
        """
        Creates a new AzureCloudContainerConnection object.

        :param storage_account_url: The storage account location of the requested
          resource.
        :param cred_manager: The credential manager used to authenticate to the
          FHIR server.
        """
        self.__storage_account_url = storage_account_url
        self.__cred_manager = cred_manager

    def _get_container_client(self, container_url: str) -> ContainerClient:
        """
        Obtains a client connected to an Azure storage container by
        utilizing the first valid credentials Azure can find. For
        more information on the order in which the credentials are
        checked, see the Azure documentation:
        https://docs.microsoft.com/en-us/azure/developer/python/sdk/authentication-overview#sequence-of-authentication-methods-when-using-defaultazurecredential

        :param container_url: The url at which to access the container.
        :return: The Azure `ContainerClient`.
        """
        creds = self.cred_manager.get_credential_object()
        return ContainerClient.from_container_url(container_url, credential=creds)

    def download_object(
        self,
        container_name: str,
        filename: str,
        encoding: Literal[None, str] = "UTF-8",
    ) -> Literal[str, bytes]:
        """
        Downloads a blob from storage and returns it as a string or bytes.

        :param container_name: The name of the container containing object to download.
        :param filename: The location of the file within Azure blob storage.
        :param encoding: The encoding applied to the downloaded content. If set to None
            then the blob contents will be returned as bytes. Default: UTF-8
        :return: A character blob as a string from the given container and filename.
        """
        container_location = f"{self.storage_account_url}/{container_name}"
        container_client = self._get_container_client(container_location)
        blob_client = container_client.get_blob_client(filename)

        downloader = blob_client.download_blob(encoding=encoding)

        blob_contents = downloader.readall()

        return blob_contents

    def upload_object(
        self,
        message: Union[str, dict],
        container_name: str,
        filename: str,
    ) -> None:
        """
        Uploads the content of a given message to Azure blob storage.
        The message can be passed either as a raw string or as JSON.

        :param message: The contents of a message, encoded either as a
          string or a JSON-formatted dictionary.
        :param container_name: The name of the target container for upload.
        :param filename: The location of file to upload within Azure blob storage.
        """
        container_location = f"{self.storage_account_url}/{container_name}"
        container_client = self._get_container_client(container_location)
        blob_client = container_client.get_blob_client(filename)

        if isinstance(message, str):
            blob_client.upload_blob(bytes(message, "utf-8"), overwrite=True)
        elif isinstance(message, dict):
            blob_client.upload_blob(json.dumps(message).encode("utf-8"), overwrite=True)

    def list_containers(self) -> List[str]:
        """
        Lists names for this CloudContainerConnection's containers.

        :return: A list of container names.
        """
        creds = self.cred_manager.get_credential_object()
        service_client = BlobServiceClient(
            account_url=self.storage_account_url, credential=creds
        )
        container_properties_generator = service_client.list_containers()

        container_name_list = []
        for container_propreties in container_properties_generator:
            container_name_list.append(container_propreties.name)

        return container_name_list

    def list_objects(self, container_name: str, prefix: str = "") -> List[str]:
        """
        Lists names for objects within a container.

        :param container_name: The name of the container to look for objects.
        :param prefix: Filter the objects returned to filenames beginning
          with this value.
        :return: A list of names for objects in given container.
        """
        container_location = f"{self.storage_account_url}/{container_name}"
        container_client = self._get_container_client(container_location)

        blob_properties_generator = container_client.list_blobs(name_starts_with=prefix)

        blob_name_list = []
        for blob_propreties in blob_properties_generator:
            blob_name_list.append(blob_propreties.name)

        return blob_name_list

    def blob_exists(self, container_name: str, filename: str) -> bool:
        """
        Check if a blob exists within a container given its name and the name of the
        container.

        :param container_name: The name of the container to look for the blob in.
        :param filename: The name of the blob to check the existence of.
        :param prefix: Filter the objects returned to filenames beginning
          with this value.
        :return: A boolean of true if the file exists and false if it does not.
        """

        container_location = f"{self.storage_account_url}/{container_name}"
        container_client = self._get_container_client(container_location)
        blob_client = container_client.get_blob_client(filename)
        return blob_client.exists()



from phdi.harmonization.double_metaphone import DoubleMetaphone
from phdi.harmonization.hl7 import convert_hl7_batch_messages_to_list
from phdi.harmonization.hl7 import default_hl7_value
from phdi.harmonization.hl7 import normalize_hl7_datetime
from phdi.harmonization.hl7 import normalize_hl7_datetime_segment
from phdi.harmonization.hl7 import standardize_hl7_datetimes
from phdi.harmonization.standardization import double_metaphone_string
from phdi.harmonization.standardization import standardize_birth_date
from phdi.harmonization.standardization import standardize_country_code
from phdi.harmonization.standardization import standardize_name
from phdi.harmonization.standardization import standardize_phone
from phdi.harmonization.utils import compare_strings

__all__ = (
    "standardize_hl7_datetimes",
    "normalize_hl7_datetime_segment",
    "normalize_hl7_datetime",
    "default_hl7_value",
    "convert_hl7_batch_messages_to_list",
    "standardize_country_code",
    "standardize_phone",
    "standardize_name",
    "double_metaphone_string",
    "compare_strings",
    "DoubleMetaphone",
    "standardize_birth_date",
)



from typing import List
from typing import Literal

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry


def http_request_with_retry(
    url: str,
    retry_count: int,
    request_type: Literal["GET", "POST"],
    allowed_methods: List[str],
    headers: dict,
    data: dict = None,
) -> requests.Response:
    """
    Executes an HTTP request, retrying the request if the returned HTTP status code
    is one of a specified list of codes.

    :param url: The url at which to make the HTTP request.
    :param retry_count: The number of times to retry the request, if the
      first attempt fails.
    :param request_type: The type of request to be made. Currently supports
      GET and POST.
    :param allowed_methods: The list of allowed HTTP request methods (i.e.,
      POST, PUT) for the specific URL and query.
    :param headers: JSON-type dictionary of headers to make the request with,
      including Authorization and content-type.
    :param data: The data as a JSON-formatted dictionary, used when the request
      requires data to be posted. Default: `None`
    :raises ValueError: An unsupported HTTP method (e.g., PATCH, DELETE) was passed
      to the request_type parameter.
    :return: A HTTP request response.
    """

    request_type = request_type.upper()
    if request_type not in ["GET", "POST"]:
        raise ValueError(
            f"The HTTP '{request_type}' method is not currently supported."
        )

    # Configure the settings of the 'requests' session we'll make
    # the API call with
    retry_strategy = Retry(
        total=retry_count,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=allowed_methods,
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    http = requests.Session()
    http.mount("http://", adapter)
    http.mount("https://", adapter)

    # Now, actually try to complete the API request
    # TODO: Condense this down to make a single call using
    # http.request(method=request_type, url=url, headers=headers, json=data)
    if request_type == "POST":
        response = http.post(
            url=url,
            headers=headers,
            json=data,
        )
    elif request_type == "GET":
        response = http.get(
            url=url,
            headers=headers,
        )

    return response



import json
import random
from functools import cache
from typing import Any
from typing import Callable
from typing import List
from typing import Literal
from typing import Union

import fhirpathpy

selection_criteria_types = Literal["first", "last", "random", "all"]


def apply_selection_criteria(
    value: List[Any],
    selection_criteria: selection_criteria_types,
) -> str | List:
    """
    Returns value(s), according to the selection criteria, from a given list of values
    parsed from a FHIR resource. A single string value is returned - if the selected
    value is a complex structure (list or dict), it is converted to a string.
    :param value: A list containing the values parsed from a FHIR resource.
    :param selection_criteria: A string indicating which element(s) of a list to select.
    :return: Value(s) parsed from a FHIR resource that conform to the selection
      criteria.
    """

    if selection_criteria == "first":
        value = value[0]
    elif selection_criteria == "last":
        value = value[-1]
    elif selection_criteria == "random":
        value = random.choice(value)
    elif selection_criteria == "all":
        return value
    else:
        raise ValueError(
            f'Selection criteria {selection_criteria} is not a valid option. Must be one of "first", "last", "random", or "all".'  # noqa
        )

    # Temporary hack to ensure no structured data is written using pyarrow.
    # Currently Pyarrow does not support mixing non-structured and structured data.
    # https://github.com/awslabs/aws-data-wrangler/issues/463
    # Will need to consider other methods of writing to parquet if this is an essential
    # feature.
    if isinstance(value, dict):  # pragma: no cover
        value = json.dumps(value)
    elif isinstance(value, list):
        value = ",".join(value)
    return value


def extract_value_with_resource_path(
    resource: dict,
    path: str,
    selection_criteria: Literal["first", "last", "random", "all"] = "first",
) -> Union[Any, None]:
    """
    Yields a single value from a resource based on a provided `fhir_path`.
    If the path doesn't map to an extant value in the first, returns
    `None` instead.
    :param resource: The FHIR resource to extract a value from.
    :param path: The `fhir_path` at which the value can be found in the
      resource.
    :param selection_criteria: A string dictating which value to extract,
      if multiple values exist at the path location.
    :return: The extracted value, or `None` if the value doesn't exist.
    """
    parse_function = get_fhirpathpy_parser(path)
    value = parse_function(resource)
    if len(value) == 0:
        return None
    else:
        value = apply_selection_criteria(value, selection_criteria)
        return value


def find_entries_by_resource_type(bundle: dict, resource_type: str) -> List[dict]:
    """
    Collect all entries of a specific type in a bundle of FHIR data and
    return references to them in a list.

    :param bundle: The FHIR bundle to search for resource entries.
    :param resource_type: The type of FHIR resource to find.
    :return: A list holding all entries of the requested resource type that were
      found in the input bundle.
    """
    return [
        entry
        for entry in bundle.get("entry", [])
        if entry.get("resource", {}).get("resourceType", "") == resource_type
    ]


def get_field(
    resource: dict,
    field: str,
    index: int = 1,
    use: str = None,
    require_use: bool = True,
) -> Any:
    """
    Finds an instance of the specified field in a given FHIR- formatted JSON dictionary.
    Optionally, a particular "use" of a field can be provided such that only instances
    with that purpose are considered. For example, find the name for a patient that has
    a "use" of "official". "Use" here refers to the FHIR-based usage of classifying a
    value's purpose. If no instance of a field with the requested use case can be found,
    instead return a specified default value for the field.

    :param resource: A FHIR-formatted resource.
    :param field: The field to extract.
    :param index: The nth element of the field to return. If the index is greater than
      the number of elements in the field, the last element will be returned. If the
      index is less than 1, the first element will be returned. Default: 1.
    :param use: The 'use' the field must have to qualify for selection. Default: None.
    :param require_use: If True and no elements of the specified field have that
      use, none will be returned. If False and no elements of the specified field have
      that use, the nth element as indicated by the index parameter will be returned.
      This parameter is ignored if no use is specified. Default: True.
    :return: The first instance of the field value matching the desired
      use, or a default field value if a match couldn't be found.
    """
    # TODO revisit the `default_field` logic, and confirm this is the best way to handle
    # choosing a default
    if field == "":
        raise ValueError("The field parameter must be a defined, non-empty string.")
    if use == "":
        raise ValueError(
            "The use parameter should be a defined, non-empty string. If you don't want to include a use, set the parameter to None."  # noqa
        )
    if field not in resource:
        raise KeyError(f"This resource does not contain a field called {field}.")

    elements = resource.get(field, [])
    if use is not None:
        elements_with_use = [item for item in elements if item.get("use") == use]
        if len(elements_with_use) == 0 and require_use:
            return None
        if len(elements_with_use) > 0:
            elements = elements_with_use

    # min(...) ensures index <= len(elements) and the -1 shifts back to 0-index
    # max(...) ensures the index is not negative
    index = max(min(index, len(elements)) - 1, 0)

    return elements[index] if len(elements) > 0 else None


@cache
def get_fhirpathpy_parser(fhirpath_expression: str) -> Callable:
    """
    Accepts a FHIRPath expression, and returns a callable function
    which returns the evaluated value at fhirpath_expression for
    a specified FHIR resource.
    :param fhirpath_expression: The FHIRPath expression to evaluate.
    :return: A function that, when called passing in a FHIR resource,
      will return value at `fhirpath_expression`.
    """
    return fhirpathpy.compile(fhirpath_expression)


def get_one_line_address(address: dict) -> str:
    """
    Extracts a one-line string representation of an address from a
    JSON dictionary holding address information.

    :param address: The FHIR-formatted address.
    :return: A one-line string representation of an address.
    """
    if len(address) == 0:
        return ""
    raw_one_line = " ".join(address.get("line", []))
    raw_one_line += f" {address.get('city', '')}, {address.get('state', '')}"
    if address.get("postalCode", ""):
        raw_one_line += f" {address['postalCode']}"
    return raw_one_line
